@@ -18,9 +18,9 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <openssl/md5.h>
 #include <assert.h>
 
+#include "md5.h"
 #include "server.h"
 
 #define CH_RING struct _ch_ring
@@ -78,12 +78,60 @@ fnv1a_hashpos(const char *key, const char *end)
 }
 
 /**
- * Qsort comparator for ch_ring_entry structs on pos.
+ * Sort comparator for ch_ring_entry structs on pos, ip and instance.
  */
 static int
-entrycmp(const void *l, const void *r)
+entrycmp_carbon(const void *l, const void *r)
 {
-	return ((ch_ring_entry *)l)->pos - ((ch_ring_entry *)r)->pos;
+	ch_ring_entry *ch_l = (ch_ring_entry *)l;
+	ch_ring_entry *ch_r = (ch_ring_entry *)r;
+
+	if (ch_l->pos != ch_r->pos)
+		return ch_l->pos - ch_r->pos;
+
+#ifndef CH_CMP_V40_BEHAVIOUR
+	{
+		int d = strcmp(server_ip(ch_l->server), server_ip(ch_r->server));
+		char *i_l, *i_r;
+		if (d != 0)
+			return d;
+		i_l = server_instance(ch_l->server);
+		i_r = server_instance(ch_r->server);
+		if (i_l == NULL && i_r == NULL)
+			return 0;
+		if (i_l == NULL)
+			return 1;
+		if (i_r == NULL)
+			return -1;
+		return strcmp(i_l, i_r);
+	}
+#endif
+
+	return 0;
+}
+
+/**
+ * Sort comparator for ch_ring_entry structs on pos, ip and port.
+ */
+static int
+entrycmp_fnv1a(const void *l, const void *r)
+{
+	ch_ring_entry *ch_l = (ch_ring_entry *)l;
+	ch_ring_entry *ch_r = (ch_ring_entry *)r;
+
+	if (ch_l->pos != ch_r->pos)
+		return ch_l->pos - ch_r->pos;
+
+#ifndef CH_CMP_V40_BEHAVIOUR
+	{
+		int d = strcmp(server_ip(ch_l->server), server_ip(ch_r->server));
+		if (d != 0)
+			return d;
+		return server_port(ch_l->server) - server_port(ch_r->server);
+	}
+#endif
+
+	return 0;
 }
 
 ch_ring *
@@ -114,6 +162,8 @@ ch_addnode(ch_ring *ring, server *s)
 	int i;
 	char buf[256];
 	ch_ring_entry *entries;
+	char *instance = server_instance(s);
+	int (*cmp)(const void *, const void *) = NULL;
 
 	if (ring == NULL)
 		return NULL;
@@ -126,7 +176,6 @@ ch_addnode(ch_ring *ring, server *s)
 	switch (ring->type) {
 		case CARBON:
 			for (i = 0; i < ring->hash_replicas; i++) {
-				char *instance = server_instance(s);
 				/* this format is actually Python's tuple format that is
 				 * used in serialised form as input for the hash */
 				snprintf(buf, sizeof(buf), "('%s', %s%s%s):%d",
@@ -135,33 +184,41 @@ ch_addnode(ch_ring *ring, server *s)
 						instance == NULL ? "None" : instance,
 						instance == NULL ? "" : "'",
 						i);
-				/* TODO:
+				/* carbon upstream committed:
 				 * https://github.com/graphite-project/carbon/commit/024f9e67ca47619438951c59154c0dec0b0518c7
-				 * Question is how harmful the collision is -- it will probably
-				 * change the location of some metrics */
+				 * this makes sure no collissions exist on pos, however,
+				 * at the expense of being agnostic to the input order,
+				 * therefore that change isn't implemented here, see
+				 * https://github.com/grobian/carbon-c-relay/issues/84 */
 				entries[i].pos = carbon_hashpos(buf, buf + strlen(buf));
 				entries[i].server = s;
 				entries[i].next = NULL;
 				entries[i].malloced = 0;
 			}
+			cmp = *entrycmp_carbon;
 			break;
 		case FNV1a:
 			for (i = 0; i < ring->hash_replicas; i++) {
 				/* take all server info into account, such that
 				 * different port numbers for the same hosts will work
-				 * (unlike CARBON) */
-				snprintf(buf, sizeof(buf), "%d-%s:%u",
-						i, server_ip(s), server_port(s));
+				 * (unlike CARBON), unless we got a full overrride */
+				if (instance == NULL) {
+					snprintf(buf, sizeof(buf), "%d-%s:%u",
+							i, server_ip(s), server_port(s));
+				} else {
+					snprintf(buf, sizeof(buf), "%d-%s", i, instance);
+				}
 				entries[i].pos = fnv1a_hashpos(buf, buf + strlen(buf));
 				entries[i].server = s;
 				entries[i].next = NULL;
 				entries[i].malloced = 0;
 			}
+			cmp = *entrycmp_fnv1a;
 			break;
 	}
 
 	/* sort to allow merge joins later down the road */
-	qsort(entries, ring->hash_replicas, sizeof(ch_ring_entry), *entrycmp);
+	qsort(entries, ring->hash_replicas, sizeof(ch_ring_entry), cmp);
 	entries[0].malloced = 1;
 
 	if (ring->entries == NULL) {
@@ -175,7 +232,7 @@ ch_addnode(ch_ring *ring, server *s)
 		last = NULL;
 		assert(ring->hash_replicas > 0);
 		for (w = ring->entries; w != NULL && i < ring->hash_replicas; ) {
-			if (w->pos < entries[i].pos) {
+			if (cmp(&w->pos, &entries[i].pos) <= 0) {
 				last = w;
 				w = w->next;
 			} else {
@@ -256,6 +313,45 @@ ch_get_nodes(
 	}
 }
 
+void
+ch_printhashring(ch_ring *ring, FILE *f)
+{
+	ch_ring_entry *w;
+	char column = 0;
+	char srvbuf[21];
+
+	for (w = ring->entries; w != NULL; w = w->next) {
+		snprintf(srvbuf, sizeof(srvbuf), "%s:%d%s%s",
+				server_ip(w->server),
+				server_port(w->server),
+				server_instance(w->server) ? "=" : "",
+				server_instance(w->server) ? server_instance(w->server) : "");
+		fprintf(f, "%5d@%-20s", w->pos, srvbuf);
+		if (column < 2) {
+			fprintf(f, " ");
+			column++;
+		} else {
+			fprintf(f, "\n");
+			column = 0;
+		}
+	}
+}
+
+unsigned short
+ch_gethashpos(ch_ring *ring, const char *key, const char *end)
+{
+	switch (ring->type) {
+		case CARBON:
+			return carbon_hashpos(key, end);
+		case FNV1a:
+			return fnv1a_hashpos(key, end);
+		default:
+			assert(0);  /* this shouldn't happen */
+	}
+
+	return 0;  /* pacify compiler */
+}
+
 /**
  * Frees the ring structure and its added nodes.
  */
