@@ -18,13 +18,13 @@
 #ifndef HAVE_RELAY_H
 #define HAVE_RELAY_H 1
 
-#define VERSION "0.40"
+#define VERSION "0.44"
 
 #define METRIC_BUFSIZ 8192
 
-enum rmode { NORMAL, DEBUG, SUBMISSION, TEST };
+enum rmode { NORMAL, DEBUG, SUBMISSION, TEST, DEBUGTEST };
 
-typedef enum { CON_TCP, CON_UDP, CON_PIPE } serv_ctype;
+typedef enum { CON_TCP, CON_UDP, CON_PIPE, CON_FILE } serv_ctype;
 
 extern char relay_hostname[];
 extern enum rmode mode;
