@@ -36,6 +36,7 @@ enum clusttype {
 	BLACKHOLE,  /* /dev/null-like destination */
 	GROUP,      /* pseudo type to create a matching tree */
 	FORWARD,
+	FILELOG,    /* like forward, write metric to file */
 	CARBON_CH,  /* room for a better/different hash definition */
 	FNV1A_CH,   /* FNV1a-based consistent-hash */
 	ANYOF,      /* FNV1a-based hash, but with backup by others */
@@ -210,6 +211,10 @@ determine_if_regex(route *r, char *pat, int flags)
  *     (forward | any_of [useall] | failover | (carbon|fnv1a)_ch [replication (count)])
  *         (ip:port[=instance] [proto (tcp | udp)] ...)
  *     ;
+ * cluster (name)
+ *     log [ip]
+ *         (/path/to/file ...)
+ *     ;
  * match (* | regex)
  *     send to (cluster | blackhole)
  *     [stop]
@@ -221,6 +226,7 @@ determine_if_regex(route *r, char *pat, int flags)
  *         (regex[ regex ...])
  *     every (interval) seconds
  *     expire after (expiration) seconds
+ *     [timestamp at (start | middle | end) of bucket]
  *     compute (sum | count | max | min | average) write to
  *         (metric)
  *     [compute ...]
@@ -254,8 +260,11 @@ router_readconfig(cluster **clret, route **rret,
 	route *r = NULL;
 	route *topr = NULL;
 	struct addrinfo *saddrs;
+	char matchcatchallfound = 0;
 
-	if ((cnf = fopen(path, "r")) == NULL || stat(path, &st) == -1)
+	if (stat(path, &st) == -1)
+		return 0;
+	if ((cnf = fopen(path, "r")) == NULL)
 		return 0;
 	buf = malloc(st.st_size + 1);
 	while ((len = fread(buf + len, 1, st.st_size - len, cnf)) != 0)
@@ -383,6 +392,18 @@ router_readconfig(cluster **clret, route **rret,
 				}
 				cl->type = FAILOVER;
 				cl->members.anyof = NULL;
+			} else if (strncmp(p, "file", 4) == 0 && isspace(*(p + 4))) {
+				p += 5;
+
+				for (; *p != '\0' && isspace(*p); p++)
+					;
+				if ((cl = cl->next = malloc(sizeof(cluster))) == NULL) {
+					logerr("malloc failed in cluster file\n");
+					free(buf);
+					return 0;
+				}
+				cl->type = FILELOG;
+				cl->members.forward = NULL;
 			} else {
 				char *type = p;
 				for (; *p != '\0' && !isspace(*p); p++)
@@ -413,9 +434,12 @@ router_readconfig(cluster **clret, route **rret,
 				struct addrinfo *next = NULL;
 				char hnbuf[256];
 
-				for (; *p != '\0' && !isspace(*p) && *p != ';'; p++)
-					if (*p == ':')
+				for (; *p != '\0' && !isspace(*p) && *p != ';'; p++) {
+					if (*p == ':' && inst == NULL)
 						lastcolon = p;
+					if (*p == '=' && inst == NULL)
+						inst = p;
+				}
 
 				if (*p == '\0') {
 					logerr("unexpected end of file at '%s' "
@@ -428,14 +452,10 @@ router_readconfig(cluster **clret, route **rret,
 				termchr = *p;
 				*p = '\0';
 
-				if (cl->type == CARBON_CH) {
-					/* parse optional "=instance" bit */
-					for (inst = p - 1; inst > ip; inst--) {
-						if (*inst == '=') {
-							*inst = '\0';
-							p = inst++;
-							break;
-						}
+				if (cl->type == CARBON_CH || cl->type == FNV1A_CH) {
+					if (inst != NULL) {
+						*inst = '\0';
+						p = inst++;
 					}
 					if (inst == ip)
 						inst = NULL;
@@ -503,36 +523,45 @@ router_readconfig(cluster **clret, route **rret,
 					}
 				}
 
-				/* resolve host/IP */
-				memset(&hint, 0, sizeof(hint));
+				if (cl->type != FILELOG) {
+					/* resolve host/IP */
+					memset(&hint, 0, sizeof(hint));
 
-				hint.ai_family = PF_UNSPEC;
-				hint.ai_socktype = *proto == 'u' ? SOCK_DGRAM : SOCK_STREAM;
-				hint.ai_protocol = *proto == 'u' ? IPPROTO_UDP : IPPROTO_TCP;
-				hint.ai_flags = AI_NUMERICSERV;
-				snprintf(sport, sizeof(sport), "%u", port);  /* for default */
+					hint.ai_family = PF_UNSPEC;
+					hint.ai_socktype = *proto == 'u' ? SOCK_DGRAM : SOCK_STREAM;
+					hint.ai_protocol = *proto == 'u' ? IPPROTO_UDP : IPPROTO_TCP;
+					hint.ai_flags = AI_NUMERICSERV;
+					snprintf(sport, sizeof(sport), "%u", port);  /* for default */
 
-				if ((err = getaddrinfo(ip, sport, &hint, &saddrs)) != 0) {
-					logerr("failed to resolve server %s:%s (%s) "
-							"for cluster %s: %s\n",
-							ip, sport, proto, name, gai_strerror(err));
-					free(cl);
-					free(buf);
-					return 0;
-				}
+					if ((err = getaddrinfo(ip, sport, &hint, &saddrs)) != 0) {
+						logerr("failed to resolve server %s:%s (%s) "
+								"for cluster %s: %s\n",
+								ip, sport, proto, name, gai_strerror(err));
+						free(cl);
+						free(buf);
+						return 0;
+					}
 
-				if (!useall && saddrs->ai_next != NULL) {
-					/* take first result only */
-					freeaddrinfo(saddrs->ai_next);
-					saddrs->ai_next = NULL;
+					if (!useall && saddrs->ai_next != NULL) {
+						/* take first result only */
+						freeaddrinfo(saddrs->ai_next);
+						saddrs->ai_next = NULL;
+					}
+				} else {
+					/* TODO: try to create/append to file */
+
+					proto = "file";
+					saddrs = (void *)1;
 				}
 
 				walk = saddrs;
 				while (walk != NULL) {
 					/* disconnect from the rest to avoid double
 					 * frees by freeaddrinfo() in server_destroy() */
-					next = walk->ai_next;
-					walk->ai_next = NULL;
+					if (walk != (void *)1) {
+						next = walk->ai_next;
+						walk->ai_next = NULL;
+					}
 
 					if (useall) {
 						/* unfold whatever we resolved, for human
@@ -556,7 +585,9 @@ router_readconfig(cluster **clret, route **rret,
 					}
 
 					newserver = server_new(ip, (unsigned short)port,
-							*proto == 'u' ? CON_UDP : CON_TCP, walk,
+							*proto == 'f' ? CON_FILE :
+							*proto == 'u' ? CON_UDP : CON_TCP,
+							walk == (void *)1 ? NULL : walk,
 							queuesize, batchsize);
 					if (newserver == NULL) {
 						logerr("failed to add server %s:%d (%s) "
@@ -582,7 +613,8 @@ router_readconfig(cluster **clret, route **rret,
 							return 0;
 						}
 						w->next = NULL;
-						if (cl->type == CARBON_CH && inst != NULL)
+						if ((cl->type == CARBON_CH || cl->type == FNV1A_CH)
+								&& inst != NULL)
 							server_set_instance(newserver, inst);
 						w->server = newserver;
 						cl->members.ch->ring = ch_addnode(
@@ -598,7 +630,8 @@ router_readconfig(cluster **clret, route **rret,
 						}
 					} else if (cl->type == FORWARD ||
 							cl->type == ANYOF ||
-							cl->type == FAILOVER)
+							cl->type == FAILOVER ||
+							cl->type == FILELOG)
 					{
 						if (w == NULL) {
 							w = malloc(sizeof(servers));
@@ -608,7 +641,9 @@ router_readconfig(cluster **clret, route **rret,
 						if (w == NULL) {
 							logerr("malloc failed for %s %s\n",
 									cl->type == FORWARD ? "forward" :
-									cl->type == ANYOF ? "any_of" : "failover",
+									cl->type == ANYOF ? "any_of" :
+									cl->type == FAILOVER ? "failover" :
+									"file",
 									ip);
 							free(cl);
 							free(buf);
@@ -616,18 +651,19 @@ router_readconfig(cluster **clret, route **rret,
 						}
 						w->next = NULL;
 						w->server = newserver;
-						if (cl->type == FORWARD && cl->members.forward == NULL)
+						if ((cl->type == FORWARD || cl->type == FILELOG)
+									&& cl->members.forward == NULL)
 							cl->members.forward = w;
-						if ((cl->type == ANYOF || cl->type == FAILOVER) &&
-								cl->members.anyof == NULL)
-						{
-							cl->members.anyof = malloc(sizeof(serverlist));
-							cl->members.anyof->count = 0;
-							cl->members.anyof->servers = NULL;
-							cl->members.anyof->list = w;
+						if (cl->type == ANYOF || cl->type == FAILOVER) {
+							if (cl->members.anyof == NULL) {
+								cl->members.anyof = malloc(sizeof(serverlist));
+								cl->members.anyof->count = 1;
+								cl->members.anyof->servers = NULL;
+								cl->members.anyof->list = w;
+							} else {
+								cl->members.anyof->count++;
+							}
 						}
-						if (cl->type == ANYOF || cl->type == FAILOVER)
-							cl->members.anyof->count++;
 					}
 
 					walk = next;
@@ -773,12 +809,21 @@ router_readconfig(cluster **clret, route **rret,
 			}
 			r->stop = w->type == BLACKHOLE ? 1 : stop;
 			r->next = NULL;
+
+			if (matchcatchallfound) {
+				logerr("warning: match %s will never be matched "
+						"due to preceeding match * ... stop\n",
+						r->pattern == NULL ? "*" : r->pattern);
+			}
+			if (r->matchtype == MATCHALL && r->stop)
+				matchcatchallfound = 1;
 		} else if (strncmp(p, "aggregate", 9) == 0 && isspace(*(p + 9))) {
 			/* aggregation rule */
 			char *type;
 			char *pat;
 			char *interval;
 			char *expire;
+			enum _aggr_timestamp tswhen = TS_END;
 			cluster *w;
 			int err;
 
@@ -895,8 +940,56 @@ router_readconfig(cluster **clret, route **rret,
 			for (; *p != '\0' && isspace(*p); p++)
 				;
 
+			/* optional timestamp bit */
+			if (strncmp(p, "timestamp", 9) == 0 && isspace(*(p + 9))) {
+				p += 10;
+				for (; *p != '\0' && isspace(*p); p++)
+					;
+				if (strncmp(p, "at", 2) != 0 || !isspace(*(p + 2))) {
+					logerr("expected 'at' after 'timestamp'\n");
+					free(buf);
+					return 0;
+				}
+				p += 3;
+				for (; *p != '\0' && isspace(*p); p++)
+					;
+				if (strncmp(p, "start", 5) == 0 && isspace(*(p + 5))) {
+					p += 6;
+					tswhen = TS_START;
+				} else if (strncmp(p, "middle", 6) == 0 && isspace(*(p + 6))) {
+					p += 7;
+					tswhen = TS_MIDDLE;
+				} else if (strncmp(p, "end", 3) == 0 && isspace(*(p + 3))) {
+					p += 4;
+					tswhen = TS_END;
+				} else {
+					logerr("expected 'start', 'middle' or 'end' "
+							"after 'timestamp at'\n");
+					free(buf);
+					return 0;
+				}
+				for (; *p != '\0' && isspace(*p); p++)
+					;
+				if (strncmp(p, "of", 2) != 0 || !isspace(*(p + 2))) {
+					logerr("expected 'of' after 'timestamp at ...'\n");
+					free(buf);
+					return 0;
+				}
+				p += 3;
+				for (; *p != '\0' && isspace(*p); p++)
+					;
+				if (strncmp(p, "bucket", 6) != 0 || !isspace(*(p + 6))) {
+					logerr("expected 'bucket' after 'timestamp at ... of'\n");
+					free(buf);
+					return 0;
+				}
+				p += 7;
+				for (; *p != '\0' && isspace(*p); p++)
+					;
+			}
+
 			w->members.aggregation =
-				aggregator_new(atoi(interval), atoi(expire));
+				aggregator_new(atoi(interval), atoi(expire), tswhen);
 			if (w->members.aggregation == NULL) {
 				logerr("invalid interval (%s) or expire (%s)\n",
 						interval, expire);
@@ -959,6 +1052,12 @@ router_readconfig(cluster **clret, route **rret,
 					;
 			} while (*p != ';');
 			p++;
+
+			if (matchcatchallfound) {
+				logerr("warning: aggregate %s will never be matched "
+						"due to preceeding match * ... stop\n",
+						r->pattern == NULL ? "*" : r->pattern);
+			}
 		} else if (strncmp(p, "rewrite", 7) == 0 && isspace(*(p + 7))) {
 			/* rewrite rule */
 			char *pat;
@@ -1038,6 +1137,12 @@ router_readconfig(cluster **clret, route **rret,
 			r->dest = cl;
 			r->stop = 0;
 			r->next = NULL;
+
+			if (matchcatchallfound) {
+				logerr("warning: rewrite %s will never be matched "
+						"due to preceeding match * ... stop\n",
+						r->pattern == NULL ? "*" : r->pattern);
+			}
 		} else {
 			/* garbage? */
 			logerr("garbage in config: %s\n", p);
@@ -1338,7 +1443,7 @@ router_getservers(cluster *clusters)
 	for (c = clusters; c != NULL; c = c->next) {
 		if (c->type == BLACKHOLE || c->type == REWRITE)
 			continue;
-		if (c->type == FORWARD) {
+		if (c->type == FORWARD || c->type == FILELOG) {
 			for (s = c->members.forward; s != NULL; s = s->next)
 				add_server(s->server);
 		} else if (c->type == ANYOF || c->type == FAILOVER) {
@@ -1360,7 +1465,7 @@ router_getservers(cluster *clusters)
  * thousands.
  */
 void
-router_printconfig(FILE *f, char all, cluster *clusters, route *routes)
+router_printconfig(FILE *f, char mode, cluster *clusters, route *routes)
 {
 	cluster *c;
 	route *r;
@@ -1378,6 +1483,11 @@ router_printconfig(FILE *f, char all, cluster *clusters, route *routes)
 			for (s = c->members.forward; s != NULL; s = s->next)
 				fprintf(f, "        %s:%d%s\n",
 						server_ip(s->server), server_port(s->server), PPROTO);
+		} else if (c->type == FILELOG) {
+			fprintf(f, "    file\n");
+			for (s = c->members.forward; s != NULL; s = s->next)
+				fprintf(f, "        %s\n",
+						server_ip(s->server));
 		} else if (c->type == ANYOF || c->type == FAILOVER) {
 			fprintf(f, "    %s\n", c->type == ANYOF ? "any_of" : "failover");
 			for (s = c->members.anyof->list; s != NULL; s = s->next)
@@ -1395,6 +1505,12 @@ router_printconfig(FILE *f, char all, cluster *clusters, route *routes)
 						PPROTO);
 		}
 		fprintf(f, "    ;\n");
+		if (mode & 2) {
+			if (c->type == CARBON_CH || c->type == FNV1A_CH) {
+				fprintf(f, "# hash ring for %s follows\n", c->name);
+				ch_printhashring(c->members.ch->ring, f);
+			}
+		}
 	}
 	fprintf(f, "\n");
 	for (r = routes; r != NULL; r = r->next) {
@@ -1402,7 +1518,7 @@ router_printconfig(FILE *f, char all, cluster *clusters, route *routes)
 			cluster *aggr = r->dest;
 			struct _aggr_computes *ac;
 
-			if (!all)
+			if (!(mode & 1))
 				continue;
 
 			fprintf(f, "aggregate\n");
@@ -1411,9 +1527,14 @@ router_printconfig(FILE *f, char all, cluster *clusters, route *routes)
 			} while (r->next != NULL && r->next->dest == aggr
 					&& (r = r->next) != NULL);
 			fprintf(f, "    every %u seconds\n"
-					"    expire after %u seconds\n",
+					"    expire after %u seconds\n"
+					"    timestamp at %s of bucket\n",
 					aggr->members.aggregation->interval,
-					aggr->members.aggregation->expire);
+					aggr->members.aggregation->expire,
+					aggr->members.aggregation->tswhen == TS_START ? "start" :
+					aggr->members.aggregation->tswhen == TS_MIDDLE ? "middle" :
+					aggr->members.aggregation->tswhen == TS_END ? "end" :
+					"<unknown>");
 			for (ac = aggr->members.aggregation->computes; ac != NULL; ac = ac->next)
 				fprintf(f, "    compute %s write to\n"
 						"        %s\n",
@@ -1485,6 +1606,7 @@ router_free(cluster *clusters, route *routes)
 				free(clusters->members.ch);
 				break;
 			case FORWARD:
+			case FILELOG:
 			case BLACKHOLE:
 				while (clusters->members.forward) {
 					server_shutdown(clusters->members.forward->server);
@@ -1710,7 +1832,7 @@ router_route_intern(
 #define failif(RETLEN, WANTLEN) \
 	if (WANTLEN > RETLEN) { \
 		logerr("router_route: out of destination slots, " \
-				"increase CONN_DESTS_SIZE in dispatcher.c\n"); \
+				"increase CONN_DESTS_SIZE in router.h\n"); \
 		return 1; \
 	}
 
@@ -1731,6 +1853,7 @@ router_route_intern(
 					break;
 			}
 			/* indirection */
+			assert(q != NULL);
 			if (*q == '\0')
 				stop = router_route_intern(
 						ret,
@@ -1746,6 +1869,7 @@ router_route_intern(
 				case BLACKHOLE: {
 					/* maybe just record we're dropping this metric? */
 				}	break;
+				case FILELOG:
 				case FORWARD: {
 					/* simple case, no logic necessary */
 					servers *s;
@@ -2015,6 +2139,11 @@ router_test_intern(char *metric, char *firstspace, route *routes)
 					fprintf(stdout, "    %s_ch(%s)\n", 
 							w->dest->type == FNV1A_CH ? "fnv1a" : "carbon",
 							w->dest->name);
+					if (mode == DEBUGTEST) {
+						fprintf(stdout, "        hash_pos(%d)\n",
+								ch_gethashpos(w->dest->members.ch->ring,
+									metric, firstspace));
+					}
 					ch_get_nodes(dst,
 							w->dest->members.ch->ring,
 							w->dest->members.ch->repl_factor,
