@@ -24,6 +24,7 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/stat.h>
 #include <netdb.h>
 #include <arpa/inet.h>
 #include <sys/resource.h>
@@ -207,6 +208,17 @@ server_queuereader(void *d)
 					self->failure += self->failure >= FAIL_WAIT_TIME ? 0 : 1;
 					continue;
 				}
+			} else if (self->ctype == CON_FILE) {
+				if ((self->fd = open(self->ip,
+								O_WRONLY | O_APPEND | O_CREAT,
+								S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0)
+				{
+					if (!self->failure)
+						logerr("failed to open file '%s': %s\n",
+								self->ip, strerror(errno));
+					self->failure += self->failure >= FAIL_WAIT_TIME ? 0 : 1;
+					continue;
+				}
 			} else {
 				int ret;
 				int args;
@@ -690,3 +702,15 @@ server_get_queue_len(server *s)
 		return 0;
 	return queue_len(s->queue);
 }
+
+/**
+ * Returns the allocated size of the queue backing metrics waiting to be
+ * sent.
+ */
+inline size_t
+server_get_queue_size(server *s)
+{
+	if (s == NULL)
+		return 0;
+	return queue_size(s->queue);
+}
